#!/usr/bin/perl
########################################################################
# SSH Command for SSH Server - Version 1.0
# (c) Georg Swoboda, 2009
########################################################################
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#########################################################################


use POSIX qw(mkfifo);
use Getopt::Std;
use Fcntl ':flock';
use IO::Handle;


##############################################
# DEFAULT USER/PASS (to hide it from ps list)
##############################################
### remove comment to enable ### my $default_user="_DEFAULT_USER";
### remove comment to enable ### my $default_pass="_DEFAULT_PASS";


########################################
# FIFO file
########################################
$FIFO="/tmp/.nagiosssh";
$FIFO_PATH="/tmp/nagiosssh/";

# TIMEOUT 
my $cmd_timeout;
my $fifo_timeout; 
my $login_timeout;
my $t_start=time();

# SAVE CURRENT ACTION
my $action="";

#local $SIG{INT} = "SigInt";
local $SIG{ALRM} = sub {  
                         print "check timed out while $action\n"; 
                         if (-p $AFIFO) {
                           unlink $AFIFO;
                         }
                         exit(3); };

# check if FIFO_PATH exists
unless(-d $FIFO_PATH){
    print "error $FIFO_PATH does not exist, is the server running ?\n";
    exit(3);
}

%options=();
getopts("hsdH:C:T:L:F:U:P:",\%options);
if ((not defined $options{H}) && (not $options{s} eq "1"))
{
 print "Error: no host given\n";
 &printHelp();
}
if ((not defined $options{C}) && (not $options{s} eq "1"))
{
 print "Error: no command given\n";
 &printHelp();
}
if (not defined $options{T})
{
 $cmd_timeout=20; # default command timeout is 15sec
}
else
{
 $cmd_timeout=$options{T};
}

if (not defined $options{L})
{
 $login_timeout=15; # default login timeout is 5sec
}
else
{
 $login_timeout=$options{L};
}

if (not defined $options{F})
{
 $fifo_timeout=3; # default fifo timeout is 2sec
}
else
{
 $fifo_timeout=$options{F};
}
if (defined $options{h})
{
 &printHelp();
}


if (!defined $default_pass)
{
 $pass=$options{P};
}
else
{
  $pass=$default_pass;
}
if (!defined $default_user)
{
 $user=$options{U};
}
else
{
 $user=$default_user;
}

if (($pass eq "")&& (not $options{s} eq "1"))
{
  print "Error: no pass given\n";
  &printHelp();
}

if (($user eq "") && (not $options{s} eq "1"))
{
  print "Error: no user given\n";
  &printHelp();
}

$host=$options{H};
$cmd=$options{C};

if (defined $options{d})
{
 $|=1;
 print "\n";
 print "\t Host: \t\t'$host'\n";
 print "\t Cmd: \t\t'$cmd'\n";
 print "\t User: \t\t'$user'\n";
 print "\t Pass: \t\t'$pass'\n";
 print "\t Login Timeout: '$login_timeout'\n";
 print "\t Cmd Timeout: \t'$cmd_timeout'\n";
 print "\t FIFO Timeout: \t'$fifo_timeout'\n";
 print "\n";
}

# make a status query
if (defined $options{s})
{
 $host="status-query";
}

#################################################
# FIFO -> SERVER
################################################# 
if (-p $FIFO)
{
 ##################################################
 # ANSWER FIFO
 ##################################################
 $AFIFO=$FIFO_PATH.$host.".".$$;
 unless (-p $AFIFO) {
            unlink $AFIFO;
            require POSIX;
            POSIX::mkfifo($AFIFO, 0777)
            or die "can't mkfifo $AFIFO: $!";
        }

 if (defined $options{d})
 {
  print ">> sending command to ssh_server.pl ";
 }
 eval {
  alarm $fifo_timeout;
  $action="flock(FIFO)";
  flock(FIFO,LOCK_EX);
  $action="open(FIFO)";
  open (FIFO, ">$FIFO") || die "can't write $FIFO: $!";
  FIFO->autoflush(1);
  $action="print(FIFO)";
  print FIFO "$host|$user|$pass|$t_start|$cmd_timeout|$login_timeout|$fifo_timeout|$$|$cmd\n";
  close FIFO;
 };
}
else
{
 print "(server.pl not running ?)\n";
 if (-p $AFIFO)
 {
  unlink $AFIFO;
 }
 exit(3);
}

if (defined $options{d})
{
  print " [DONE]\n";
}

#################################################
# CHILD(SERVER) -> COMMAND
#################################################

# calc timeout that we wait at max
my $timeout=$login_timeout+$fifo_timeout+$fifo_timeout+$cmd_timeout;

if (defined $options{d})
{
  print ">> waiting max '$timeout' secs for answer\n\n";
}

eval {
 alarm $timeout;
 $action="open(AFIFO)($timeout secs) '$AFIFO'";
 if (-p $AFIFO)
 { 
  open (AFIFO, "$AFIFO") || die "can't read $AFIFO: $!";
  $status=0;
  $answer="";
  $action="read(AFIFO)";
  while (<AFIFO>)
        {
         $l=$_;
         if (!$status)
         {
          ($host,$tstamp,$exitcode)=split(/\|/,$l);
          $status=1; 
         }
         else
         {
          $answer=$answer.$l;
         }
        }
  close AFIFO;
 }
};

if (-p $AFIFO)
{
    unlink $AFIFO;
}

print $answer;
my $t_end=time();
my $t_diff=$t_end-$t_start;
print "Execution Time: $t_diff Sec(s)\n";
exit $exitcode;

sub printHelp()
{
 print "./command.pl [-d] [-s] -U <user> -P <pass> -H <host> -C <command> -T <ssh-cmd-timeout> -L <ssh-login-timeout> -F <fifo-timeout>\n\n";
 exit 3;
}

######################################################################
# CRTL^C - remove incoming fifo and all answer fifos
######################################################################
sub SigInt()
{
 print "SigInt()\n";
 close(FIFO);
 close(AFIFO);
 unlink $AFIFO; 
 exit(0);
}

