#!/usr/bin/perl
########################################################################
# SSH Server - Version 1.0
# (c) Georg Swoboda, 2009
########################################################################
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#########################################################################

use POSIX qw(mkfifo);
use POSIX ":sys_wait_h";
use IO::Handle;
use IO::Select;
use Net::SSH::Perl;
use Net::SSH::Perl::Constants qw( :msg );
use Fcntl ':flock';

########################################################################
# FIFO file for command.pl
########################################################################
my $FIFO="/tmp/.nagiosssh";
my $FIFO_PATH="/tmp/nagiosssh/";


# SSH debug flag
my $ssh_debug=0;

# where the servers stores its $$   
my $pid_file="/root/nagiosssh/server.pid"; 
# CHILD's answer fifo 
my $AFIFO;
# TIMEOUTS
my $afifo_timeout=3;
my $login_timeout=5;

# parent PID
my $parent_pid;

# childs SSH session
my $ssh_session=undef;
# known workers
my %pidpool;   # pids of known workers

# pipe's() to workers
my $PARENT_RDR;
my $CHILD_WTR;
my $CHILD_RDR;
my $PARENT_WTR;

# select object for childs
my $s=IO::Select->new();

# error/debug vars
my $cur_host, $cur_cmd, $cur_action, $cur_cmd_pid, $cur_timeout, $cur_user;

# statics
$|=1;
$parent_pid=$$;
umask 000;

# check if dirs are there

unless(-d $FIFO_PATH){
    print "(creating $FIFO_PATH dir)\n";
    mkdir $FIFO_PATH or die("unable to mkdir $FIFO_PATH\n");
}

# remove the command.pl fifo when starting
unlink $FIFO;
&ServerPid();


######################################################################
# signal handler
######################################################################
local $SIG{INT} = "SigInt";
local $SIG{ALRM} = sub {  
       my $what="CHILD[".$$."]";
       my $err_str=&get_timestamp."TIMEOUT in $what after $cur_timeout sec(s) while '$cur_action' user='$cur_user' host='$cur_host' cmd='$cur_cmd' pid=$cur_cmd_pid ]";
       print "\033[1;31m$err_str \033[0m\n";  
       &sendAnswer($cur_host,$cur_cmd_pid,3,$err_str,1);
       # kill the child, and cleanup
       &DieChild($parent_pid,$cur_host,$cur_user);
       #####################################################
       # the parent isnt allowed to exit ;-)
       print ">> parent wont exit\n";
       return undef;
       };


####################################################
# create the command.pl fifo
#
unless (-p $FIFO) {
 unlink $FIFO;
 require POSIX;
 POSIX::mkfifo($FIFO, 0777)
 or die "can't mkfifo $FIFO: $!\n";
}


####################################################
# open the fifo from command.pl
print &get_timestamp."Server is ready, waiting for command.pl requests\n\n";
while (1)
{
       ####################################################
       # next line blocks until there's a reader
       open(FIFO, "$FIFO") or die "cant open fifo $FIFO: $!\n";
       while (<FIFO>) 
       {
         $l=$_; 
         ($host,$user,$pass,$t_start,$cmd_timeout,$login_timeout,$fifo_timeout,$cmd_pid,$cmd)=split(/\|/,$l);
         chop($cmd);
         # check if status is requested or cmd
         if ($host eq "status-query")
         {
          &ServerStatus(); # generate status report 
         }
         else
         {
           if (!defined($pidpool{$host." ".$user}))
           {
             #################################################
             # unknown host, create worker
             &forkWorker($host,$user);
           }
           ###################################################
           # tell Worker what to do
           &tellWorker($cmd,$user,$pass,$t_start,$cmd_timeout,$login_timeout,$fifo_timeout,$cmd_pid,$CHILD_WTR{$host." ".$user},$host);
        }
       }
       close(FIFO);
}
# we never reach this

######################################################################
# send command to Worker
######################################################################
sub tellWorker()
{
 my $cmd=@_[0];
 my $user=@_[1];
 my $pass=@_[2];
 my $t_start=@_[3];
 my $cmd_timeout=@_[4];
 my $login_timeout=@_[5];
 my $fifo_timeout=@_[6];
 my $cmd_pid=@_[7];
 my $F=@_[8];
 my $host=@_[9];

 local $SIG{PIPE} = sub {
                          #
                          # there should be a worker, but there was none
                          # so we can assume it exited with an error
                          # so lets recreate the worker and tell him what to do
                          #
                          close($F); 
                          delete $pidpool{$host." ".$user};
                          &forkWorker($host,$user);
                          $F=$CHILD_WTR{$host." ".$user};
                          print $F $cmd."|".$user."|".$pass."|".$t_start."|".$cmd_timeout."|".$login_timeout."|".$fifo_timeout."|".$cmd_pid."\n";
                        };

 print $F $cmd."|".$user."|".$pass."|".$t_start."|".$cmd_timeout."|".$login_timeout."|".$fifo_timeout."|".$cmd_pid."\n";
 return;
}

######################################################################
# create a pipe(), fork the Child (runWorker) 
# and dispatch the first command
######################################################################
sub forkWorker()
{
 my $host=@_[0];
 my $user=@_[1];

 ###########################################################
 # setup the 2way communication to the runWorker process

 pipe($PARENT_RDR{$host." ".$user}, $CHILD_WTR{$host." ".$user});               
 pipe($CHILD_RDR{$host." ".$user}, $PARENT_WTR{$host." ".$user});              
 $CHILD_WTR{$host." ".$user}->autoflush(1);
 $PARENT_WTR{$host." ".$user}->autoflush(1);

 $flags |= O_NONBLOCK; # Add non-blocking to the flags
 fcntl($CHILD_WTR{$host." ".$user}, F_SETFL, $flags);

 ##########################################################
 # create child, and send first command
 if ($pid = fork)
 {
   $pidpool{$host." ".$user}["pid"]=$pid;  # register child with %pidpool
   close $PARENT_RDR{$host." ".$user};
   close $PARENT_WTR{$host." ".$user};
   waitpid(-1,WNOHANG);
 }
 else # child
 {
   close $CHILD_RDR{$host." ".$user};
   close $CHILD_WTR{$host." ".$user};
   &runWorker($host,$PARENT_RDR{$host." ".$user},$PARENT_WTR{$host." ".$user});
 }
}

######################################################################
# Mainloop for the Child       
# waits for requests via PARENT_RDR
######################################################################
sub runWorker()
{
 my $host=$_[0];
 my $P_RDR=$_[1];
 my $P_WTR=$_[2];
 my $pid=$$;

 my $cmd_timeout,$login_timeout,$fifo_timeout;
 my $cmd,$cmd_pid,$t_start;

# print ">>>Worker($pid) spawned, waiting for tasks\n";
 while (<$P_RDR>)
 { 
#  print " >>> Worker got task\n";
  ($cmd,$user,$pass,$t_start,$cmd_timeout,$login_timeout,$fifo_timeout,$cmd_pid)=split(/\|/,$_);
  chop($cmd_pid);
  runCommand($host,$user,$pass,$t_start,$cmd_timeout,$login_timeout,$fifo_timeout,$cmd,$cmd_pid);
 }
}

######################################################################
# login to host, return $ssh handle to be used in the child
######################################################################
sub loginSSH()
{
  my $host=$_[0];
  my $user=$_[1];
  my $pass=$_[2];
  my $timeout=$_[3];

  &updateState($host,"",$cmd,$cmd_pid,"ssh->login()",$timeout);
  eval {
   alarm $timeout;
   $ssh=Net::SSH::Perl->new($host, options=>["PubkeyAuthentication no","KeepAlive no"], debug=>$ssh_debug, protocol=>'2,1' ,port=>22);
   &updateAction("ssh->login()","User=$user");
   $ssh->login($user,$pass);
   alarm 0;
  };
  return $ssh;
}

######################################################################
# check if the child has an active ssh_session
# if not, create one
######################################################################
sub fetchSSH()
{
  my $host=$_[0];
  my $user=$_[1];
  my $pass=$_[2];
  my $timeout=$_[3];

  if (!defined($ssh_session))
  {
   $ssh_session=&loginSSH($host,$user,$pass,$timeout);
   if (not defined $ssh_session)
   {
    sshSessionDead();
   }
   # add sock to select group
   $s->add($ssh_session->sock);
  }
}

######################################################################
# set program status, used for debug
######################################################################
sub updateState()
{
  $cur_host=$_[0];
  $cur_msg=$_[1];
  $cur_cmd=$_[2];
  $cur_cmd_pid=$_[3];
  $cur_action=$_[4];
  $cur_timeout=$_[5];
}

sub updateAction()
{
  $cur_action=$_[0];
  $cur_msg=$_[1];
}

######################################################################
# execute command, received via FIFO
######################################################################
sub runCommand()
{
   my $host=$_[0];
   my $user=$_[1];
   my $pass=$_[2];
   my $t_start=$_[3];
   my $cmd_timeout=$_[4];
   my $login_timeout=$_[5];
   my $fifo_timeout=$_[6];
   my $cmd=$_[7]; 
   my $cmd_pid=$_[8];

   # set cur_user
   $cur_user=$user;

   &updateState($host,"",$cmd,$cmd_pid,"",$login_timeout);   
   &fetchSSH($host,$user,$pass,$login_timeout);

   ########################################################################
   # check if the ssh sock is still writeable
   if (defined($ssh_session))
   {
     if ($s->can_write(0))
     {
       &updateState($host,"",$cmd,$cmd_pid,"ssh->cmd()",$cmd_timeout);
       ######################################################################
       # signal handler if ssh connection died
       local $SIG{PIPE} = sub {
                            &sshSessionDead();
                           };
       eval
       {
         alarm $cmd_timeout;
         ($stdout, $stderr, $exit)=$ssh_session->cmd($cmd);
         alarm 0;
       };
       &sendAnswer($host,$cmd_pid,$exit,$stdout,$fifo_timeout);
      }
      else
      {
       ####################################################################
       # ssh sock not writeable
       &updateState($host,"",$cmd,$cmd_pid,"IO::Select",0);
       &sshSessionDead();
      }
   } 
   else 
   {
     ####################################################################
     # ssh sock unknown 
     &sshSessionDead();
   }
   ######################################################################
   # calc time
   my $t_diff=time()-$t_start;
   if ($t_diff > ($cmd_timeout/2))
   {
    print "Slow Command: $t_diff sec(s) '$cur_host' '$cur_cmd'\n";
   }
}


######################################################################
# send message back to command.pl
######################################################################
sub sendAnswer()
{
  my $host=$_[0];
  my $cmd_pid=$_[1];
  my $exitcode=$_[2];
  my $answer=$_[3];
  my $fifo_timeout=$_[4];

  # set answer fifo
  $AFIFO=$FIFO_PATH.$host.".".$cmd_pid;

  &updateAction("sendAnswer",$answer);
  if (-p $AFIFO) 
  {
     #################################################################
     # if the fifo is dead (command.pl gone) cleanup the mess
     local $SIG{ALRM} = sub {
                             my $err_str=&get_timestamp."AFIFO-TIMEOUT in CHILD[".$$."] after $fifo_timeout sec(s) while '$cur_action' host='$cur_host' cmd='$cur_cmd' pid=$cur_cmd_pid ]";
                             print "\033[1;31m$err_str \033[0m\n";
                             close($AFIFO);
                             if (-p $AFIFO)
                             {
                               unlink $AFIFO;
                             }
                            };
     eval 
     {
       alarm $fifo_timeout;
       open (AFIFO, ">$AFIFO") || die "can't write $AFIFO: $!";
       print AFIFO "$host|$tstamp|$exitcode\n$answer\n";
       close AFIFO;
       alarm 0;
     };
  }
  else
  {
   print &get_timestamp."no valid AFIFO ($$) ($AFIFO)\n";
  }
  return;
}

######################################################################
# CRTL^C - remove incoming fifo and all answer fifos
######################################################################
sub SigInt()
{
# print "SigInt()\n";
 if ($$ == $parent_pid)
 {
  close(FIFO);
  unlink $FIFO;  # server fifo
 } 
 exit(0);
}

######################################################################
# ssh session died, inform command.pl, and kill child
######################################################################
sub sshSessionDead()
{
 # for some reason the ssh session died, kill the child
 my $what="CHILD[".$$."]";
 my $err_str=&get_timestamp."SSH SESSION DEAD in $what while '$cur_action' user='$cur_user' host='$cur_host' cmd='$cur_cmd' pid=$cur_cmd_pid ]";
 print "\033[1;31m$err_str \033[0m\n";
 &sendAnswer($cur_host,$cur_cmd_pid,3,$err_str,1);
 &DieChild($parent_pid,$cur_host);
}

######################################################################
# let child die, and close all connections
######################################################################
sub DieChild()
{
 my $parent_pid=$_[0];
 my $cur_host=$_[1];

 if ($$ != $parent_pid)
 {
      ####################################################
      # child is dead, closing all socks
      eval 
      {
       $s->remove($ssh_session->sock);
      };
      close($PARENT_RDR{$cur_host." ".$cur_user});
      close($PARENT_WTR{$cur_host." ".$cur_user});
      close($AFIFO);
      exit(-1);
      return undef; die "timeout\n";
 }
 else
 {
  print "DieChild() in parent, that should not happen !\n";
 }
}

######################################################################
# dump $$ to $pid_file
######################################################################
sub ServerPid()
{
 open(F,">$pid_file");
 print F $$;
 close(F);
}
 
#####################################################################
# gen timestamp
#####################################################################
sub get_timestamp {
   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
   if ($mon < 10) { $mon = "0$mon"; }
   if ($hour < 10) { $hour = "0$hour"; }
   if ($min < 10) { $min = "0$min"; }
   if ($sec < 10) { $sec = "0$sec"; }
   $year=$year+1900;
   return "[".$year . ' ' . $mon . '/' . $mday . ' - ' . $hour . ':' . $min . ':' . $sec."] -\t";
}

sub ServerStatus()
{
  my $report="",$c=0;
  
  foreach $w (keys %pidpool)
  {
   my $s_host,$s_user,$s_pid;
   my $s_stime,$s_ltime;
   ($s_host,$s_user)=split(" ",$w);
   $s_pid=$pidpool{$w}["pid"];

   if (!waitpid($s_pid,WNOHANG))
   {
     $report=$report."$s_host\t$s_user\t$s_pid\n";
     $c++;
   }
 }
 $report="======== Process Status ========\n\n".$report;
$report=$report."\nActive SSH Connections: $c\n";
 &sendAnswer($host,$cmd_pid,0,$report,$fifo_timeout);
}
